import typing
from functools import partial
from typing import Any, Optional

import chex
import jax

from esquilax import utils


def amap(f: typing.Callable) -> typing.Callable:
    """
    Decorator that maps a function over agent state data

    Maps an update function over state data,
    broadcasting model parameters, and assigning unique
    random keys. The size of the output data is the same
    as the inputs, i.e. it produces a result for each agent.

    Inputs and outputs can be PyTrees of arrays
    where each array entry represent an agent. See
    below for examples.

    Examples
    --------

    .. testsetup:: amap

       import esquilax
       import jax

    .. testcode:: amap

       def foo(p, x):
           return p + x

       a = jax.numpy.arange(5)

       result = esquilax.transforms.amap(foo)(2, a)
       # [2, 3, 4, 5, 6]

    .. doctest:: amap
       :hide:

       >>> result.tolist()
       [2, 3, 4, 5, 6]

    It can also be used as a decortor. Arguments can
    also be PyTrees, and parameters ``None`` if unused

    .. testcode:: amap

       @esquilax.transforms.amap
       def foo(_, x):
           return x[0] + x[1]

       a = (jax.numpy.arange(5), jax.numpy.arange(5))

       result = foo(None, a)
       # result = [0, 2, 4, 6, 8]

    .. doctest:: amap
       :hide:

       >>> result.tolist()
       [0, 2, 4, 6, 8]

    Arguments can also multidimensional (as long as
    the first axis is the number of agents), and outputs
    can be PyTrees

    .. testcode:: amap

       @esquilax.transforms.amap
       def foo(_, x):
           # Returns a tuple
           return x[0], x[1]

       a = jax.numpy.arange(10).reshape(5, 2)

       foo(None, a)
       # ([0, 2, 4, 6, 8], [1, 3, 5, 7, 9])

    .. doctest:: amap
       :hide:

       >> foo(None, a)
       ([0, 2, 4, 6, 8], [1, 3, 5, 7, 9])

    JAX random keys can be passed to the wrapped function
    by including the ``key`` keyword argument

    .. testcode:: amap

       def foo(_, x, *, key):
           return x + jax.random.choice(key, 100, ())

       k = jax.random.PRNGKey(101)
       result = esquilax.transforms.amap(foo)(None, a, key=k)

    Parameters
    ----------
    f
        Update function that should have the signature

        .. code-block:: python

           def f(params, x, **static_kwargs):
               ...
               return y

        where the arguments are:

        - ``params``: Parameters (shared across the map)
        - ``x``: State data to map over.
        - ``**static_kwargs``: Any values required at compile
          time by JAX can be passed as keyword arguments.

        Random keys can be passed to the wrapped function
        by including the ``key`` keyword argument.
    """

    keyword_args = utils.functions.get_keyword_args(f)
    has_key, keyword_args = utils.functions.has_key_keyword(keyword_args)

    @partial(jax.jit, static_argnames=keyword_args)
    def _self(
        params: Any,
        x: chex.ArrayTree,
        key: Optional[chex.PRNGKey] = None,
        **static_kwargs,
    ) -> chex.ArrayTree:
        chex.assert_tree_has_only_ndarrays(x)
        n = utils.functions.get_size(x)

        if has_key:
            assert key is not None, "Expected keyword argument 'key'"
            keys = jax.random.split(key, n)
            results = jax.vmap(
                lambda k, p, a: f(p, a, key=k, **static_kwargs), in_axes=(0, None, 0)
            )(keys, params, x)
        else:
            assert key is None, "Received unexpected 'key' keyword argument"
            results = jax.vmap(partial(f, **static_kwargs), in_axes=(None, 0))(
                params, x
            )

        return results

    return _self



from functools import partial
from math import prod
from typing import Any, Callable, Optional, Tuple

import chex
import jax
import jax.numpy as jnp

from esquilax import utils
from esquilax.reductions import Reduction


def _sort_agents(
    dims: Tuple[int, int], co_ords: chex.Array, agents: chex.Array
) -> Tuple[chex.Array, chex.Array, chex.Array, chex.Array, chex.ArrayTree]:
    idxs = co_ords[:, 0] * dims[1] + co_ords[:, 1]
    sort_idxs = jnp.argsort(idxs)
    _, bins = utils.graph.index_bins(idxs, prod(dims))
    sorted_co_ords = co_ords[sort_idxs]
    sorted_idxs = idxs[sort_idxs]
    sorted_agents = jax.tree_util.tree_map(lambda y: y[sort_idxs], agents)

    return sorted_co_ords, sort_idxs, bins, sorted_idxs, sorted_agents


def _check_arguments(
    idxs: chex.Array,
    idxs_b: Optional[chex.Array],
    agents_a: Optional[chex.ArrayTree],
    agents_b: Optional[chex.ArrayTree],
) -> None:
    assert (
        idxs.ndim == 2 and idxs.shape[1] == 2
    ), f"idxs argument should be an array of 2d indices, got shape {idxs.shape}"

    if idxs_b is not None:
        assert idxs_b.ndim == 2 and idxs_b.shape[1] == 2, (
            "idxs_b argument should be an array of "
            f"2d coordinates, got shape {idxs_b.shape}"
        )
        n_b = idxs_b.shape[:1]
    else:
        n_b = idxs.shape[:1]

    if agents_a is not None:
        chex.assert_tree_has_only_ndarrays(agents_a)
        chex.assert_tree_shape_prefix(agents_a, idxs.shape[:1])

    if agents_b is not None:
        chex.assert_tree_has_only_ndarrays(agents_b)
        chex.assert_tree_shape_prefix(agents_b, n_b)


def grid(
    f: Callable,
    *,
    reduction: Reduction,
    dims: Tuple[int, int],
    include_self: bool = False,
    topology: str = "moore",
) -> Callable:
    """
    Apply a function to agents based on a grid neighbourhood

    Applies an interaction function between agents based
    on their proximity on a 2-dimensional grid.

    .. warning::

       This implementation assumes the grid is wrapped at its
       boundaries.

    Examples
    --------

    This example subdivides the space into 2 cells along
    each dimension (i.e. there are 4 cells total),
    then sums contributions from neighbours, excluding the
    active agent

    .. testsetup:: grid

       import esquilax
       import jax
       import jax.numpy as jnp
       from functools import partial

    .. testcode:: grid

       def foo(p, a, b):
           return p + a + b

       x = jnp.array([[0, 0], [0, 1], [2, 2]])
       a = jnp.arange(3)

       result = esquilax.transforms.grid(
           foo,
           reduction=esquilax.reductions.add(dtype=int),
           dims=(4, 4),
           include_self=False,
       )(
           2, a, a, co_ords=x
       )
       # result = [3, 3, 0]

    .. doctest:: grid
       :hide:

       >>> result
       Array([3, 3, 0], dtype=int32)

    in this case the first two agents are next
    to each other on the grid, but the third
    has no direct neighbours.

    The transform can also be used as a decorator using
    :py:meth:`functools.partial`. Arguments and return values
    can be PyTrees or multidimensional arrays. Arguments can
    also be ``None`` if not used

    .. testcode:: grid

       tuple_reduce = esquilax.reductions.Reduction(
           fn=(jnp.add, jnp.add), id=(0, 0),
       )

       @partial(
           esquilax.transforms.grid,
           reduction=tuple_reduce,
           dims=(4, 4),
           include_self=False,
       )
       def foo(p, _, b):
           return p + b[0], p + b[1]

       x = jnp.array([[0, 0], [0, 1], [2, 2]])
       a = jnp.arange(6).reshape(3, 2)

       result = foo(2, None, a, co_ords=x)
       # result = ([4, 2, 0], [5, 3, 0])

    .. doctest:: grid

       >>> result
       (Array([4, 2, 0], dtype=int32), Array([5, 3, 0], dtype=int32))

    You can also pass different agent types (i.e. two sets
    of agents with different positions) using the ``co_ords_b``
    keyword argument

    .. testcode:: grid

       def foo(p, a, b):
           return p + a + b

       xa = jnp.array([[0, 0], [2, 2]])
       xb = jnp.array([[0, 0], [2, 2]])
       a = jnp.arange(2)
       b = 2 + a

       result = esquilax.transforms.grid(
           foo,
           reduction=esquilax.reductions.add(dtype=int),
           dims=(4, 4),
           include_self=False,
       )(
           2, a, b, co_ords=xa, co_ords_b=xb
       )
       # result = [4, 6]

    .. doctest:: grid
       :hide:

       >>> result
       Array([4, 6], dtype=int32)

    .. testcode:: grid

       def foo(p, a, b, *, key):
           return jax.random.choice(key, 100, ())

       k = jax.random.PRNGKey(101)
       result = esquilax.transforms.grid(
           foo,
           reduction=esquilax.reductions.add(dtype=int),
           dims=(4, 4),
           include_self=False,
       )(
           None, None, None, co_ords=x, key=k
       )

    Parameters
    ----------
    f
        Interaction to apply to in-proximity pairs, should
        have the signature

        .. code-block:: python

           def f(
               params: Any,
               a: Any,
               b: Any,
               **static_kwargs,
           ):
               ...
               return x

        where

        - ``params``: Parameters broadcast over all interactions
        - ``a``: Start agent in the interaction
        - ``b``: End agent in the interaction
        - ``**static_kwargs``: Any arguments required at compile
          time by JAX can be passed as keyword arguments.

        Random keys can be passed to the wrapped function by
        including the ``key`` keyword argument.
    reduction
        Binary monoidal reduction.
    dims
        Number of cells along each dimension
    include_self
        if ``True`` each agent will include itself in the
        gathered values.
    topology
        Topology of cells, default ``"moore"``. Since cells
        interact with their neighbours, topologies with
        fewer neighbours can increase performance at the
        cost of fidelity. Should be one of ``"same-cell"``,
        ``"von-neumann"`` or ``"moore"``.
    """
    offsets = utils.space.get_neighbours_offsets(topology)
    keyword_args = utils.functions.get_keyword_args(f)
    has_key, keyword_args = utils.functions.has_key_keyword(keyword_args)

    @partial(jax.jit, static_argnames=keyword_args)
    def _grid(
        params: Any,
        agents_a: Any,
        agents_b: Any,
        *,
        co_ords: chex.Array,
        co_ords_b: Optional[chex.Array] = None,
        key: Optional[chex.PRNGKey] = None,
        **static_kwargs,
    ) -> Any:
        utils.functions.check_key(has_key, key)
        _check_arguments(co_ords, co_ords_b, agents_a, agents_b)
        same_types = co_ords_b is None

        (
            co_ords_a,
            sort_idxs_a,
            bins_a,
            sorted_idxs_a,
            sorted_agents_a,
        ) = _sort_agents(dims, co_ords, agents_a)

        if same_types:
            bins_b = bins_a
            sorted_agents_b = jax.tree_util.tree_map(lambda y: y[sort_idxs_a], agents_b)
        else:
            _, _, bins_b, _, sorted_agents_b = _sort_agents(dims, co_ords_b, agents_b)

        def cell(k: Optional[chex.PRNGKey], i: int, bin_range: chex.Array) -> Any:
            agent_a = jax.tree_util.tree_map(lambda y: y[i], sorted_agents_a)

            def interact(
                j: int, carry: Tuple[Optional[chex.PRNGKey], Any]
            ) -> Tuple[Optional[chex.PRNGKey], Any]:
                _k, _r = carry
                agent_b = jax.tree_util.tree_map(lambda z: z[j], sorted_agents_b)

                if has_key:
                    _k, fk = jax.random.split(_k, 2)
                    r = f(params, agent_a, agent_b, key=fk, **static_kwargs)
                else:
                    r = f(params, agent_a, agent_b, **static_kwargs)

                r = jax.tree_util.tree_map(
                    lambda red, a, b: red(a, b), reduction.fn, _r, r
                )
                return _k, r

            if (not same_types) or include_self:
                _, _results = jax.lax.fori_loop(
                    bin_range[0], bin_range[1], interact, (k, reduction.id)
                )
            else:
                k, _results = jax.lax.fori_loop(
                    bin_range[0],
                    jnp.minimum(i, bin_range[1]),
                    interact,
                    (k, reduction.id),
                )
                _, _results = jax.lax.fori_loop(
                    jnp.maximum(i + 1, bin_range[0]),
                    bin_range[1],
                    interact,
                    (k, _results),
                )

            return _results

        def agent_reduce(
            k: Optional[chex.PRNGKey], i: chex.Numeric, _co_ords: chex.Array
        ):
            nbs = utils.space.neighbour_indices(_co_ords, offsets, dims)
            nb_bins = bins_b[nbs]

            if has_key:
                _keys = jax.random.split(k, nbs.shape[0])
                _results = jax.vmap(cell, in_axes=(0, None, 0))(_keys, i, nb_bins)
            else:
                _results = jax.vmap(cell, in_axes=(None, None, 0))(None, i, nb_bins)

            def red(a, _, c):
                return jnp.frompyfunc(c, 2, 1).reduce(a)

            return jax.tree_util.tree_map(red, _results, reduction.id, reduction.fn)

        n_agents = co_ords.shape[0]

        if has_key:
            keys = jax.random.split(key, n_agents)
            results = jax.vmap(agent_reduce, in_axes=(0, 0, 0))(
                keys, jnp.arange(n_agents), co_ords_a
            )
        else:
            results = jax.vmap(agent_reduce, in_axes=(None, 0, 0))(
                None, jnp.arange(n_agents), co_ords_a
            )

        inv_sort = jnp.argsort(sort_idxs_a)
        results = jax.tree_util.tree_map(lambda y: y[inv_sort], results)

        return results

    return _grid


def grid_local(
    f: Callable,
    *,
    topology: str = "moore",
) -> Callable:
    """
    Apply a function to an agent local cells

    Applies an interaction function between an agent and
    its surrounding cells. The transform slices the
    cell saround the agent, and passes them
    to the mapped function.

    .. warning::

       This implementation assumes the grid is wrapped at its
       boundaries.

    Examples
    --------

    This example subdivides the space into 3 cells along
    each dimension (i.e. there are 9 cells total),
    then applies the function to an agent at ``[1, 1]``

    .. testsetup:: grid

       import esquilax
       import jax
       import jax.numpy as jnp
       from functools import partial

    .. testcode:: grid

       def foo(p, _, grid):
           # grid here is an array of values
           return p + grid

       x = jnp.array([[1, 1]])
       grid = jnp.arange(9).reshape(3, 3)

       result = esquilax.transforms.grid_local(
           foo
       )(
           2, None, grid, co_ords=x
       )
       # result = [[ 2,  3,  4,  5,  6,  7,  8,  9, 10]]

    .. doctest:: grid
       :hide:

       >>> result
       Array([[ 2,  3,  4,  5,  6,  7,  8,  9, 10]], dtype=int32)

    The transform can also be used as a decorator using
    :py:meth:`functools.partial`. Arguments and return values
    can be PyTrees or multidimensional arrays. Arguments can
    also be ``None`` if not used

    .. testcode:: grid

       @partial(
           esquilax.transforms.grid_local,
           topology="von-neumann",
       )
       def foo(p, _, g):
           return p[0] + g[0], p[1] + g[1]

       x = jnp.array([[1, 1]])
       grid = jnp.arange(9).reshape(3, 3)
       grids = (grid, grid + 1)

       result = foo((1, 2), None, grids, co_ords=x)
       # result = ([[5, 6, 8, 4, 2]], [[ 7,  8, 10,  6,  4]])

    .. doctest:: grid
       :hide:

       >>> result[0]
       Array([[5, 6, 8, 4, 2]], dtype=int32)
       >>> result[1]
       Array([[ 7,  8, 10,  6,  4]], dtype=int32)

    Random keys can be passed to the wrapped function by
    including the ``key`` keyword argument

    .. testcode:: grid

       def foo(_p, _a, grid, *, key):
           # grid here is an array of values
           return grid + jax.random.choice(key, 100, ())

       k = jax.random.PRNGKey(101)
       result = esquilax.transforms.grid_local(
           foo
       )(
           2, None, grid, co_ords=x, key=k
       )

    Parameters
    ----------
    f
        Interaction to apply to in-proximity pairs, should
        have the signature

        .. code-block:: python

           def f(
               params: Any,
               a: Any,
               grid: Any,
               **static_kwargs,
           ):
               ...
               return x

        where

        - ``params``: Parameters broadcast over all interactions
        - ``a``: Agent in the interaction
        - ``grid``: Pytree of arrays gathered from local cells.
        - ``**static_kwargs``: Any arguments required at compile
          time by JAX can be passed as keyword arguments.

        JAX random keys can be passed to the wrapped argument by
        including the ``key`` keyword argument.
    topology
        Topology of the local neighbourhood, default ``"moore"``.
        The cells passed to the interaction function are gathered
        according to this topology, e.g. a moore neighbourhood
        will pass an array of size ``(9,)`` to the function.
    """
    offsets = utils.space.get_neighbours_offsets(topology)
    keyword_args = utils.functions.get_keyword_args(f)
    has_key, keyword_args = utils.functions.has_key_keyword(keyword_args)

    @partial(jax.jit, static_argnames=keyword_args)
    def _grid_local(
        params: Any,
        agents: chex.ArrayTree,
        grids: chex.ArrayTree,
        key: Optional[chex.PRNGKey] = None,
        *,
        co_ords: chex.Array,
        **static_kwargs,
    ) -> chex.ArrayTree:
        assert grids is not None, "'grids' argument should not be None"
        utils.functions.check_key(has_key, key)
        dims = jax.tree.flatten(grids)[0][0].shape[:2]
        chex.assert_tree_shape_prefix(grids, dims)
        dims = jnp.array(dims)
        _check_arguments(co_ords, None, agents, None)

        def inner(k: Optional[chex.PRNGKey], _co_ords: chex.Array, agent):
            nbs = (_co_ords[jnp.newaxis] + offsets) % dims
            grid_vals = jax.tree.map(lambda x: x[nbs[:, 0], nbs[:, 1]], grids)
            if has_key:
                result = f(params, agent, grid_vals, key=k, **static_kwargs)
            else:
                result = f(params, agent, grid_vals, **static_kwargs)
            return result

        n_agents = co_ords.shape[0]

        if has_key:
            keys = jax.random.split(key, n_agents)
            results = jax.vmap(inner, in_axes=(0, 0, 0))(keys, co_ords, agents)
        else:
            results = jax.vmap(inner, in_axes=(None, 0, 0))(None, co_ords, agents)

        return results

    return _grid_local



"""
Utilities for handling PyTrees of :py:class:`esquilax.ml.evo.strategy`
"""
from typing import Tuple

import chex
import evosax
import jax

from esquilax.typing import TypedPyTree
from esquilax.utils import tree

from .strategy import Strategy


def tree_ask(
    key: chex.PRNGKey,
    strategies: TypedPyTree[Strategy],
    evo_states: TypedPyTree[evosax.EvoState],
    evo_params: TypedPyTree[evosax.EvoParams],
) -> Tuple[TypedPyTree[evosax.EvoState], chex.ArrayTree, chex.ArrayTree]:
    """
    Call ``ask`` on PyTree of strategies, generating population samples

    Generate a new population sample of parameters, and
    also reshape those parameters for use in  a
    function/neural network.

    Parameters
    ----------
    key
        JAX random key
    strategies
        PyTree of strategies, could be a single strategy
        or a container/struct of strategies.
    evo_states
        PyTree of strategy states
    evo_params
        PyTree of strategy parameters

    Returns
    -------
    tuple
        Tuple containing:

        - Updated strategy states
        - Population samples
        - Population samples reshaped for use in
          a function/neural network

        Each will have the same tree structure as
        the input arguments.
    """
    keys = tree.key_tree_split(key, strategies, Strategy)

    def inner(strat, k, state, params):
        pop, state = strat.ask(k, state, params)
        pop_shaped = strat.reshape_params(pop)
        return state, pop, pop_shaped

    results = jax.tree.map(
        inner,
        strategies,
        keys,
        evo_states,
        evo_params,
        is_leaf=lambda x: isinstance(x, Strategy),
    )
    evo_states, population, population_shaped = tree.transpose_tree_of_tuples(
        strategies, results, 3, Strategy
    )
    return evo_states, population, population_shaped


def tree_tell(
    strategies: TypedPyTree[Strategy],
    populations: chex.ArrayTree,
    rewards: chex.ArrayTree,
    evo_states: TypedPyTree[evosax.EvoState],
    evo_params: TypedPyTree[evosax.EvoParams],
) -> Tuple[chex.ArrayTree, TypedPyTree[evosax.EvoState]]:
    """
    Call ``tell`` on a PyTree of strategies updating state

    Reshape/rescale rewards, and use this to update the
    state of the strategy.

    Parameters
    ----------
    strategies
        PyTree of strategies, could be a single strategy
        or a container/struct of strategies.
    populations
        PyTree of populations.
    rewards
        PyTree of rewards.
    evo_states
        PyTree of strategy states.
    evo_params
        PyTree of strategy parameters.

    Returns
    -------
    tuple
        Tuple containing:

        - PyTree of fitness values (i.e. the rescaled,
          reshaped rewards).
        - PyTree of updated strategy states.

        Each has the same tree structure as the arguments.
    """

    def inner(strat, pop, rew, state, params):
        f = strat.shape_rewards(pop, rew)
        state = strat.tell(pop, f, state, params)
        return f, state

    updates = jax.tree.map(
        inner,
        strategies,
        populations,
        rewards,
        evo_states,
        evo_params,
        is_leaf=lambda x: isinstance(x, Strategy),
    )
    fitness, evo_states = tree.transpose_tree_of_tuples(
        strategies, updates, 2, Strategy
    )
    return fitness, evo_states



"""
Function and data utilities
"""
import inspect
from typing import Callable, List, Optional, Tuple

import chex
import jax


def get_size(tree: chex.ArrayTree) -> int:
    """
    Get the length of arrays in a PyTree

    Also checks that arrays are all the same length

    Examples
    --------

    .. testsetup:: get_size

       import jax.numpy as jnp
       from esquilax.utils.functions import get_size

    .. doctest:: get_size

       >>> x = (jnp.arange(10), jnp.arange(10))
       >>> get_size(x)
       10

    Parameters
    ----------
    tree
        PyTree of arrays

    Raises
    ------
    AssertionError
        Raised if arrays do not have the same length.

    Returns
    -------
    int
        Length of arrays in the tree
    """
    lens = jax.tree_util.tree_map(lambda x: x.shape[0], tree)
    lens = jax.tree_util.tree_flatten(lens)[0]
    assert all([lens[0] == n for n in lens[1:]])
    return lens[0]


def get_keyword_args(f: Callable) -> List[str]:
    """
    Get keyword-only argument names from a function signature

    Examples
    --------

    .. testsetup:: get_keyword_args

       import jax.numpy as jnp
       from esquilax.utils.functions import get_keyword_args

    .. doctest:: get_keyword_args

       >>> def foo(a, b, *, c, d): pass
       >>> get_keyword_args(foo)
       ['c', 'd']

    Parameters
    ----------
    f
        Function to inspect

    Returns
    -------
    list[str]
        List of keyword-only argument names.
    """
    sig = inspect.signature(f)
    key_word_args = [
        k for k, p in sig.parameters.items() if p.kind == inspect.Parameter.KEYWORD_ONLY
    ]
    return key_word_args


def has_key_keyword(keyword_args: List[str]) -> Tuple[bool, List[str]]:
    """
    Check if list of arguments contains the "key" keyword

    Parameters
    ----------
    keyword_args
        List of keyword arguments

    Returns
    -------
    tuple[bool, list[str]]
        Flag if "key" was present, and updated list of arguments
    """
    has_key = "key" in keyword_args
    keys = [k for k in keyword_args if k != "key"]
    return has_key, keys


def check_key(has_key: bool, key: Optional[chex.PRNGKey]) -> None:
    """
    Assert if a key instance has/not been provided if expected

    Parameters
    ----------
    has_key
        Flag indicating if a wrapped function expects a JAX random key
    key
        Optional JAX random key that should be None if ``has_key`` is
        ``False`` or an instance of a JAX key otherwise.

    Returns
    -------

    """
    if has_key:
        assert key is not None, "Expected keyword argument 'key'"
    else:
        assert key is None, "Received unexpected 'key' keyword argument"



"""
Reduction interface and common implementations
"""
from typing import Callable, Sequence

import chex
import jax
import jax.numpy as jnp
from jax._src.numpy.ufuncs import ufunc

from esquilax.typing import TypedPyTree


class Reduction:
    """
    Class wrapping a reduction function and its default/identity value

    A reduction function and its associated default value. The
    reduction function and identity can also be PyTrees to
    allow reduction of multiple values in a single pass.
    """

    __slots__ = ["fn", "id"]

    def __init__(
        self, fn: TypedPyTree[Callable | ufunc], id: TypedPyTree[chex.Numeric]
    ) -> None:
        """
        Initialise a reduction

        Parameters
        ----------
        fn
            Binary reduction function.
        id
            Identity/default value of the reduction.
        """
        chex.assert_trees_all_equal_structs(
            fn, id
        ), "'fn' and 'id' PyTrees should have the same structure"
        self.fn = fn
        self.id = id


def add(shape: Sequence[int] = (), dtype: jax.typing.DTypeLike = float) -> Reduction:
    """
    Addition reduction

    Reduction that adds values, with a default value of ``0``.

    Parameters
    ----------
    shape
        Shape of reduced values.
    dtype
        Dtype of reduced values.

    Returns
    -------
    Reduction
        Initialised :py:class:`Reduction`.
    """
    return Reduction(fn=jnp.add, id=jnp.zeros(shape, dtype=dtype))


def logical_or(shape: Sequence[int] = ()) -> Reduction:
    """
    Logical or reduction

    Reduction that applies the or function to values, with a default value of ``False``.

    Parameters
    ----------
    shape
        Shape of reduced values.

    Returns
    -------
    Reduction
        Initialised :py:class:`Reduction`.
    """
    return Reduction(fn=jnp.logical_or, id=jnp.zeros(shape, dtype=bool))


def min(shape: Sequence[int] = (), dtype: jax.typing.DTypeLike = float) -> Reduction:
    """
    Minimum reduction

    Reduction that returns the minimum of values, with a default value
    of the maximum numerical values of the dtype.

    Parameters
    ----------
    shape
        Shape of reduced values.
    dtype
        Dtype of reduced values.

    Returns
    -------
    Reduction
        Initialised :py:class:`Reduction`.
    """
    val = jnp.finfo(dtype).max
    return Reduction(fn=jnp.min, id=jnp.full(shape, val, dtype=dtype))


def max(shape: Sequence[int] = (), dtype: jax.typing.DTypeLike = float) -> Reduction:
    """
    Maximum reduction

    Reduction that returns the maximum of values, with a default value
    of the minimum numerical values of the dtype.

    Parameters
    ----------
    shape
        Shape of reduced values.
    dtype
        Dtype of reduced values.

    Returns
    -------
    Reduction
        Initialised :py:class:`Reduction`.
    """
    val = jnp.finfo(dtype).min
    return Reduction(fn=jnp.max, id=jnp.full(shape, val, dtype=dtype))
